/*  Copyright (c) MediaArea.net SARL & Reto Kromer.
 *
 *  Use of this source code is governed by a BSD-style license that can
 *  be found in the License.html file in the root of the source tree.
 */

//---------------------------------------------------------------------------
#include "Lib/License/License.h"
#include "Lib/License/License_Internal.h"
#include "Lib/Utils/FileIO/Input_Base.h"
#define __STDC_WANT_LIB_EXT1__ 1
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <iomanip>
#if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__))
    #include <Shlobj.h>
    #define PathSeparator "\\"
#else
    #include <sys/stat.h>
    #define PathSeparator "/"
#endif
using namespace std;
//---------------------------------------------------------------------------

//***************************************************************************
// Secure version of some OS calls
//***************************************************************************

//---------------------------------------------------------------------------
static string getenv_string(const char* name)
{
    #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__)) // TODO: getenv_s is supported with GCC only with C++17.
        // Get size
        size_t len;
        if (getenv_s(&len, NULL, 0, name) || !len)
            return string();

        // Get value
        char* value = new char[len * sizeof(char)];
        if (getenv_s(&len, value, len, name))
            return string();

        // Create string
        string valueString(value, len);
        delete[] value;
        return valueString;
    #else
        // FIXME: this is not secure
        //struct _reent Reent;
        //return string(&Reent, getenv_s(name));
        const char* ToReturn = getenv(name);
        if (ToReturn)
            return ToReturn;
        return string();
    #endif
}

//---------------------------------------------------------------------------
static tm localtime_tm(const time_t& time)
{
    tm result;
    #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__))
        localtime_s(&result, &time);
    #else
        localtime_r(&time, &result);
    #endif

    return result;
}

//---------------------------------------------------------------------------
bool CreateDirectory (const string& Dir)
{
    #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__))
        DWORD dwAttrib = GetFileAttributesA(Dir.c_str());
        if (dwAttrib != INVALID_FILE_ATTRIBUTES && (dwAttrib & FILE_ATTRIBUTE_DIRECTORY))
            return false;
        if (CreateDirectoryA(Dir.c_str(), NULL))
            return false;
        #else
        struct stat s = {};
        if (!stat(Dir.c_str(), &s) && (s.st_mode & S_IFDIR))
            return false;
        if (!mkdir(Dir.c_str(), 0700))
            return false;
    #endif

    cerr << "Problem while creating " << Dir << " directory" << endl;
        
    return true;
}

//---------------------------------------------------------------------------
string GetLocalConfigPath (bool Create = false)
{
    string result;
    #if (defined(WIN32) || defined(_WIN32) || defined(__WIN32__))
        char szPath[MAX_PATH];
        if(SUCCEEDED(SHGetFolderPathA(NULL, CSIDL_APPDATA, NULL, 0, szPath)))
            result = szPath;
    #else
        result = getenv_string("HOME");
        if (!result.empty())
            result += '/';
        result += ".config";
    #endif

    if (!result.empty())
    {
        if (Create)
            CreateDirectory(result);
        result += PathSeparator "RAWcooked";
        if (Create)
            CreateDirectory(result);
    }
    return result;
}

//***************************************************************************
// License input
//***************************************************************************

//---------------------------------------------------------------------------
class license_input : public input_base
{
public:
    license_input() : input_base(Parser_License) {}

    license_internal*           License = nullptr;

private:
    void                        ParseBuffer();
    void                        BufferOverflow();
};

//---------------------------------------------------------------------------
void license_input::BufferOverflow()
{
}

//***************************************************************************
// Place holder for license encoding/decoding
//***************************************************************************

//---------------------------------------------------------------------------
// Place holder for license key related functions - Begin

//---------------------------------------------------------------------------
// Flags - Features
static const feature Features_Flags_List[] =
{
    feature::GeneralOptions,
    feature::InputOptions,
    feature::EncodingOptions,
    feature::MultipleTracks,
    feature::SubLicense,
};
static_assert(sizeof(Features_Flags_List) / sizeof(feature) == feature_Max, "feature license issue");

//---------------------------------------------------------------------------
// Flags - Muxers
static const muxer Muxers_Flags_List[] =
{
    muxer::Matroska,
};
static_assert(sizeof(Muxers_Flags_List) / sizeof(muxer) == muxer_Max, "muxer license issue");

//---------------------------------------------------------------------------
// Flags - Encoders
static const encoder Encoders_Flags_List[] =
{
    encoder::PCM,
    encoder::FFV1,
    encoder::FLAC,
};
static_assert(sizeof(Encoders_Flags_List) / sizeof(encoder) == encoder_Max, "encoder license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - DPX
static const dpx::flavor DPX_Flags_List[] =
{
    dpx::flavor::Raw_RGBA_10_FilledA_LE,
    dpx::flavor::Raw_RGBA_10_FilledA_BE,
    dpx::flavor::Raw_RGB_16_BE,
    dpx::flavor::Raw_RGB_16_LE,
    dpx::flavor::Raw_RGBA_16_BE,
    dpx::flavor::Raw_RGBA_16_LE,
    dpx::flavor::Raw_RGB_8,
    dpx::flavor::Raw_RGBA_8,
    dpx::flavor::Raw_RGB_10_FilledA_LE,
    dpx::flavor::Raw_RGB_10_FilledA_BE,
    dpx::flavor::Raw_RGB_12_Packed_BE,
    dpx::flavor::Raw_RGB_12_FilledA_BE,
    dpx::flavor::Raw_RGB_12_FilledA_LE,
    dpx::flavor::Raw_RGBA_12_Packed_BE,
    dpx::flavor::Raw_RGBA_12_FilledA_BE,
    dpx::flavor::Raw_RGBA_12_FilledA_LE,
    dpx::flavor::Raw_Y_8,
    dpx::flavor::Raw_Y_16_LE,
    dpx::flavor::Raw_Y_16_BE,
    dpx::flavor::Raw_Y_10_FilledA_BE,
    dpx::flavor::Raw_Y_10_FilledB_BE,
    dpx::flavor::Raw_Y_12_Packed_BE,
};
static_assert(sizeof(DPX_Flags_List) / sizeof(dpx::flavor) == dpx::flavor_Max, "DPX license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - TIFF
static const tiff::flavor TIFF_Flags_List[] =
{
    tiff::flavor::Raw_RGB_16_U_BE,
    tiff::flavor::Raw_RGB_16_U_LE,
    tiff::flavor::Raw_RGBA_16_U_LE,
    tiff::flavor::Raw_RGB_8_U,
    tiff::flavor::Raw_RGBA_8_U,
    tiff::flavor::Raw_Y_8_U,
    tiff::flavor::Raw_Y_16_U_LE,
    tiff::flavor::Raw_Y_16_U_BE,
};
static_assert(sizeof(TIFF_Flags_List) / sizeof(tiff::flavor) == tiff::flavor_Max, "TIFF license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - WAV
static const wav::flavor WAV_Flags_List[] =
{
    wav::flavor::PCM_48000_24_1_LE,
    wav::flavor::PCM_48000_24_2_LE,
    wav::flavor::PCM_48000_24_6_LE,
    wav::flavor::PCM_96000_24_1_LE,
    wav::flavor::PCM_96000_24_2_LE,
    wav::flavor::PCM_96000_24_6_LE,
    wav::flavor::PCM_48000_16_1_LE,
    wav::flavor::PCM_48000_16_2_LE,
    wav::flavor::PCM_48000_16_6_LE,
    wav::flavor::PCM_96000_16_1_LE,
    wav::flavor::PCM_96000_16_2_LE,
    wav::flavor::PCM_96000_16_6_LE,
    wav::flavor::PCM_48000_8_1_U,
    wav::flavor::PCM_48000_8_2_U,
    wav::flavor::PCM_48000_8_6_U,
    wav::flavor::PCM_96000_8_1_U,
    wav::flavor::PCM_96000_8_2_U,
    wav::flavor::PCM_96000_8_6_U,
    wav::flavor::PCM_44100_24_1_LE,
    wav::flavor::PCM_44100_24_2_LE,
    wav::flavor::PCM_44100_24_6_LE,
    wav::flavor::PCM_44100_16_1_LE,
    wav::flavor::PCM_44100_16_2_LE,
    wav::flavor::PCM_44100_16_6_LE,
    wav::flavor::PCM_44100_8_1_U,
    wav::flavor::PCM_44100_8_2_U,
    wav::flavor::PCM_44100_8_6_U,
    wav::flavor::PCM_44100_8_8_U,
    wav::flavor::PCM_44100_16_8_LE,
    wav::flavor::PCM_44100_24_8_LE,
    wav::flavor::PCM_48000_8_8_U,
    wav::flavor::PCM_48000_16_8_LE,
    wav::flavor::PCM_48000_24_8_LE,
    wav::flavor::PCM_96000_8_8_U,
    wav::flavor::PCM_96000_16_8_LE,
    wav::flavor::PCM_96000_24_8_LE,
    wav::flavor::PCM_44100_32_1_LE,
    wav::flavor::PCM_44100_32_2_LE,
    wav::flavor::PCM_44100_32_6_LE,
    wav::flavor::PCM_44100_32_8_LE,
    wav::flavor::PCM_48000_32_1_LE,
    wav::flavor::PCM_48000_32_2_LE,
    wav::flavor::PCM_48000_32_6_LE,
    wav::flavor::PCM_48000_32_8_LE,
    wav::flavor::PCM_96000_32_1_LE,
    wav::flavor::PCM_96000_32_2_LE,
    wav::flavor::PCM_96000_32_6_LE,
    wav::flavor::PCM_96000_32_8_LE,
    wav::flavor::PCM_44100_32_1_LE_F,
    wav::flavor::PCM_44100_32_2_LE_F,
    wav::flavor::PCM_44100_32_6_LE_F,
    wav::flavor::PCM_44100_32_8_LE_F,
    wav::flavor::PCM_48000_32_1_LE_F,
    wav::flavor::PCM_48000_32_2_LE_F,
    wav::flavor::PCM_48000_32_6_LE_F,
    wav::flavor::PCM_48000_32_8_LE_F,
    wav::flavor::PCM_96000_32_1_LE_F,
    wav::flavor::PCM_96000_32_2_LE_F,
    wav::flavor::PCM_96000_32_6_LE_F,
    wav::flavor::PCM_96000_32_8_LE_F,
    wav::flavor::PCM_44100_8_4_U,
    wav::flavor::PCM_44100_16_4_LE,
    wav::flavor::PCM_44100_24_4_LE,
    wav::flavor::PCM_44100_32_4_LE_F,
    wav::flavor::PCM_44100_32_4_LE,
    wav::flavor::PCM_48000_8_4_U,
    wav::flavor::PCM_48000_16_4_LE,
    wav::flavor::PCM_48000_24_4_LE,
    wav::flavor::PCM_48000_32_4_LE_F,
    wav::flavor::PCM_48000_32_4_LE,
    wav::flavor::PCM_96000_8_4_U,
    wav::flavor::PCM_96000_16_4_LE,
    wav::flavor::PCM_96000_24_4_LE,
    wav::flavor::PCM_96000_32_4_LE_F,
    wav::flavor::PCM_96000_32_4_LE,
};
static_assert(sizeof(WAV_Flags_List) / sizeof(wav::flavor) == wav::flavor_Max, "WAV license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - EXR
static const exr::flavor EXR_Flags_List[] =
{
    exr::flavor::Raw_RGB_16,
};
static_assert(sizeof(EXR_Flags_List) / sizeof(exr::flavor) == exr::flavor_Max, "EXR license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - AIFF
static const aiff::flavor AIFF_Flags_List[] =
{
    aiff::flavor::PCM_48000_24_1_BE,
    aiff::flavor::PCM_48000_24_2_BE,
    aiff::flavor::PCM_48000_24_6_BE,
    aiff::flavor::PCM_96000_24_1_BE,
    aiff::flavor::PCM_96000_24_2_BE,
    aiff::flavor::PCM_96000_24_6_BE,
    aiff::flavor::PCM_48000_16_1_BE,
    aiff::flavor::PCM_48000_16_1_LE,
    aiff::flavor::PCM_48000_16_2_BE,
    aiff::flavor::PCM_48000_16_2_LE,
    aiff::flavor::PCM_48000_16_6_BE,
    aiff::flavor::PCM_48000_16_6_LE,
    aiff::flavor::PCM_96000_16_1_BE,
    aiff::flavor::PCM_96000_16_1_LE,
    aiff::flavor::PCM_96000_16_2_BE,
    aiff::flavor::PCM_96000_16_2_LE,
    aiff::flavor::PCM_96000_16_6_BE,
    aiff::flavor::PCM_96000_16_6_LE,
    aiff::flavor::PCM_44100_8_1_S,
    aiff::flavor::PCM_44100_8_1_U,
    aiff::flavor::PCM_44100_8_2_S,
    aiff::flavor::PCM_44100_8_2_U,
    aiff::flavor::PCM_44100_8_6_S,
    aiff::flavor::PCM_44100_8_6_U,
    aiff::flavor::PCM_44100_16_1_BE,
    aiff::flavor::PCM_44100_16_1_LE,
    aiff::flavor::PCM_44100_16_2_BE,
    aiff::flavor::PCM_44100_16_2_LE,
    aiff::flavor::PCM_44100_16_6_BE,
    aiff::flavor::PCM_44100_16_6_LE,
    aiff::flavor::PCM_44100_24_1_BE,
    aiff::flavor::PCM_44100_24_2_BE,
    aiff::flavor::PCM_44100_24_6_BE,
    aiff::flavor::PCM_48000_8_1_S,
    aiff::flavor::PCM_48000_8_1_U,
    aiff::flavor::PCM_48000_8_2_S,
    aiff::flavor::PCM_48000_8_2_U,
    aiff::flavor::PCM_48000_8_6_S,
    aiff::flavor::PCM_48000_8_6_U,
    aiff::flavor::PCM_96000_8_1_S,
    aiff::flavor::PCM_96000_8_1_U,
    aiff::flavor::PCM_96000_8_2_S,
    aiff::flavor::PCM_96000_8_2_U,
    aiff::flavor::PCM_96000_8_6_S,
    aiff::flavor::PCM_96000_8_6_U,
    aiff::flavor::PCM_44100_8_8_S,
    aiff::flavor::PCM_44100_8_8_U,
    aiff::flavor::PCM_44100_16_8_BE,
    aiff::flavor::PCM_44100_16_8_LE,
    aiff::flavor::PCM_44100_24_8_BE,
    aiff::flavor::PCM_48000_8_8_S,
    aiff::flavor::PCM_48000_8_8_U,
    aiff::flavor::PCM_48000_16_8_BE,
    aiff::flavor::PCM_48000_16_8_LE,
    aiff::flavor::PCM_48000_24_8_BE,
    aiff::flavor::PCM_96000_8_8_S,
    aiff::flavor::PCM_96000_8_8_U,
    aiff::flavor::PCM_96000_16_8_BE,
    aiff::flavor::PCM_96000_16_8_LE,
    aiff::flavor::PCM_96000_24_8_BE,
    aiff::flavor::PCM_44100_32_1_BE,
    aiff::flavor::PCM_44100_32_2_BE,
    aiff::flavor::PCM_44100_32_6_BE,
    aiff::flavor::PCM_44100_32_8_BE,
    aiff::flavor::PCM_48000_32_1_BE,
    aiff::flavor::PCM_48000_32_2_BE,
    aiff::flavor::PCM_48000_32_6_BE,
    aiff::flavor::PCM_48000_32_8_BE,
    aiff::flavor::PCM_96000_32_1_BE,
    aiff::flavor::PCM_96000_32_2_BE,
    aiff::flavor::PCM_96000_32_6_BE,
    aiff::flavor::PCM_96000_32_8_BE,
    aiff::flavor::PCM_44100_8_4_S,
    aiff::flavor::PCM_44100_8_4_U,
    aiff::flavor::PCM_44100_16_4_BE,
    aiff::flavor::PCM_44100_16_4_LE,
    aiff::flavor::PCM_44100_24_4_BE,
    aiff::flavor::PCM_44100_32_4_BE,
    aiff::flavor::PCM_48000_8_4_S,
    aiff::flavor::PCM_48000_8_4_U,
    aiff::flavor::PCM_48000_16_4_BE,
    aiff::flavor::PCM_48000_16_4_LE,
    aiff::flavor::PCM_48000_24_4_BE,
    aiff::flavor::PCM_48000_32_4_BE,
    aiff::flavor::PCM_96000_8_4_S,
    aiff::flavor::PCM_96000_8_4_U,
    aiff::flavor::PCM_96000_16_4_BE,
    aiff::flavor::PCM_96000_16_4_LE,
    aiff::flavor::PCM_96000_24_4_BE,
    aiff::flavor::PCM_96000_32_4_BE,
};
static_assert(sizeof(AIFF_Flags_List) / sizeof(aiff::flavor) == aiff::flavor_Max, "AIFF license issue");

//---------------------------------------------------------------------------
// Flags - Parsers - AVI
static const avi::flavor AVI_Flags_List[] =
{
    avi::flavor::v210,
};
static_assert(sizeof(AVI_Flags_List) / sizeof(avi::flavor) == avi::flavor_Max, "AVI license issue");

//---------------------------------------------------------------------------
// License mapping
struct license_map
{
    const uint8_t               Pos;
    const uint8_t               List_Offset;
    const uint8_t* List;
};
static const license_map License_Map[] =
{
    {                             (uint8_t)-1, 0, nullptr}, // Temporary
    {                             (uint8_t)-1, 0, nullptr}, // Owner ID
    {                                       0, 0, (const uint8_t*)Features_Flags_List},
    { License_Parser_Offset + Parser_DPX     , 0, (const uint8_t*)DPX_Flags_List},
    { License_Parser_Offset + Parser_WAV     , 0, (const uint8_t*)WAV_Flags_List},
    { License_Parser_Offset + Parser_TIFF    , 0, (const uint8_t*)TIFF_Flags_List},
    { License_Parser_Offset + Parser_AIFF    , 0, (const uint8_t*)AIFF_Flags_List},
    { License_Parser_Offset + Parser_EXR     , 0, (const uint8_t*)EXR_Flags_List},
    {                                       1, 0, (const uint8_t*)Muxers_Flags_List},
    {                                       2, 0, (const uint8_t*)Encoders_Flags_List},
    {                             (uint8_t)-1, 0, nullptr},
    {                             (uint8_t)-1, 0, nullptr},
    {                             (uint8_t)-1, 0, nullptr},
    {                             (uint8_t)-1, 0, nullptr},
    { License_Parser_Offset + Parser_AIFF    , 1, (const uint8_t*)AIFF_Flags_List},
    { License_Parser_Offset + Parser_WAV     , 1, (const uint8_t*)WAV_Flags_List},
    { License_Parser_Offset + Parser_AVI     , 0, (const uint8_t*)AVI_Flags_List},
};
static const size_t Flags_Map_Size = sizeof(License_Map) / sizeof(decltype(*License_Map));

//---------------------------------------------------------------------------
static void Put_EB(uint8_t* Buffer, size_t& Offset, uint64_t Size, uint64_t Value)
{
    size_t N_l = 1;
    while (Value >> (N_l * 7))
        N_l++;
    if (Value == (1 << (N_l * 7)) - 1)
        N_l++; // This value is reserved for "unlimited"
    uint64_t N2 = Value | (((uint64_t)1) << (N_l * 7));
    while (N_l)
    {
        Buffer[Offset] = (uint8_t)(N2 >> ((N_l - 1) * 8));
        Offset++;
        if (Offset >= Size)
        {
            Offset++;
            break;
        }
        N_l--;
    }
}
// Place holder for license key related functions - End

//---------------------------------------------------------------------------
buffer license_internal::ToBuffer()
{
    buffer Buffer;

    // Place holder for license key encoding - Begin
    Buffer.Create(100);
    Buffer[0] = 0x00; // Version
    Buffer[1] = 0x00; // Checksum (filled later)
    size_t Buffer_Offset = 2;

    // Registering all
    uint64_t Values[56] = {};
    uint64_t Flags = 0;
    for (uint8_t i = 0; i < 56; i++)
    {
        switch (i)
        {
        case 0: // Temporary
            if (Date != (uint64_t)-1)
            {
                auto TimeInfo = localtime_tm(Date);
                uint64_t Date_Limit = ((TimeInfo.tm_year - 100) * 12 + TimeInfo.tm_mon) % 128;

                Flags |= ((uint64_t)1 << i);
                Values[i] = Date_Limit;
            }
            break;
        case 1: // Entity ID
            if (OwnerID)
            {
                Flags |= ((uint64_t)1 << i);
                Values[i] = OwnerID;
            }
            break;
        case 55: // SubLicensee ID
            if (SubLicenseeID)
            {
                Flags |= ((uint64_t)1 << i);
                Values[i] = SubLicenseeID;
                break;
            }
            break;
        default:
            if (i < Flags_Map_Size && License_Map[i].Pos != (uint8_t)-1) // Flags
            {
                auto Offset = License_Map[i].List_Offset * 56;
                auto Max = License_Infos[License_Map[i].Pos].Size - Offset;
                if (Max > 56)
                    Max = 56;
                for (uint8_t j = 0; j < Max; j++)
                    if (IsSupported(License_Map[i].Pos, License_Map[i].List[Offset + j]))
                    {
                        Flags |= ((uint64_t)1 << i);
                        Values[i] |= ((uint64_t)1 << j);
                    }
            }
        }
    }
    Put_EB(Buffer.Data(), Buffer_Offset, Buffer.Size(), Flags);
    for (uint8_t i = 0; i < 56; i++)
        if (Flags & ((uint64_t)1 << i))
            Put_EB(Buffer.Data(), Buffer_Offset, Buffer.Size(), Values[i]);

    // Quick and dirty scrambling: i depends on i-1, then i-1 depends on i, then pseudo random offst
    // It changes nearly all bytes if 1 flag is modified e.g. a license is update with 1 item
    for (size_t i = 3; i < Buffer_Offset; i++)
    {
        Buffer[i] -= Buffer[i - 1];
    }
    for (size_t i = Buffer_Offset - 2; i >= 2; i--)
    {
        Buffer[i] -= Buffer[i + 1];
    }
    uint8_t Scramble = 0xA5;
    for (size_t i = 2; i < Buffer_Offset; i++)
    {
        Buffer[i] -= Scramble;
        Scramble += 0xA5;
    }

    // Checksum
    uint8_t Check = 0;
    for (size_t i = 2; i < Buffer_Offset; i++)
        Check ^= Buffer[i];
    Buffer[1] = Check;

    Buffer.Resize(Buffer_Offset);
    // Place holder for license key encoding - End

    return Buffer;
}

//---------------------------------------------------------------------------
void license_input::ParseBuffer()
{
    // Place holder for license key decoding - Begin
    SetDetected();
    SetSupported();
    if (!License)
        return;

    // Temporary buffer
    buffer BufferMod;
    BufferMod.Create(Buffer.Data(), Buffer.Size());
    auto Data = BufferMod.Data();
    auto Size = BufferMod.Size();

    // Size
    if (Size < 2)
    {
        cerr << "Invalid size of license key." << endl;
        return;
    }

    // Version
    if (Data[0])
    {
        cerr << "Unsupported license key version." << endl;
        return;
    }

    // Checksum
    uint8_t Check = 0;
    for (size_t i = 2; i < Size; i++)
        Check ^= Data[i];
    if (Check != Data[1])
    {
        cerr << "Invalid license key." << endl;
        return;
    }

    // Decrambling: revert i depends on i-1, then i-1 depends on i, then pseudo random offst
    uint8_t Scramble = 0xA5;
    for (size_t i = 2; i < Size; i++)
    {
        Data[i] += Scramble;
        Scramble += 0xA5;
    }
    for (size_t i = 2; i <= Size - 2; i++)
    {
        Data[i] += Data[i + 1];
    }
    for (size_t i = Size - 1; i >= 3; i--)
    {
        Data[i] += Data[i - 1];
    }

    // Use temporary buffer
    Buffer = buffer_view(BufferMod);

    // Up to 56 56-bit values
    uint64_t Flags;
    Buffer_Offset = 2;
    Flags = Get_EB();
    for (uint8_t i = 0; i < 56; i++)
    {
        if (Flags & ((uint64_t)1 << i))
        {
            uint64_t Value = Get_EB();
            switch (i)
            {
            case 0: // Temporary
            {
                uint64_t Date_Limit;
                Date_Limit = Value;
                time_t Time;
                time(&Time);
                struct tm TimeInfo = localtime_tm(Time);
                int Now = (TimeInfo.tm_year - 100) * 12 + TimeInfo.tm_mon;

                while (Now >= 64 && Now > Date_Limit && Now - Date_Limit >= 128 - 12) // 12-month sliding so maximum temporary key duration
                    Date_Limit += 128;

                struct tm TimeInfo_Limit;
                memset(&TimeInfo_Limit, 0x00, sizeof(tm));
                TimeInfo_Limit.tm_year = 100 + (int)(Date_Limit + 1) / 12;
                TimeInfo_Limit.tm_mon = (Date_Limit + 1) % 12;
                TimeInfo_Limit.tm_mday = 1;
                time_t Date_Limit2 = mktime(&TimeInfo_Limit);
                License->Date = Date_Limit2 - TimeInfo_Limit.tm_isdst * 3600 - 1; // 1 second before the next month in order to get the last day of the previous month
                break;
            }
            case 1: // Entity ID
                License->OwnerID = Value;
                break;
            case 55: // SubLicensee ID
                License->SubLicenseeID = Value;
                break;
            default:
                if (i < Flags_Map_Size && License_Map[i].Pos != (uint8_t)-1) // Flags
                {
                    auto Offset = License_Map[i].List_Offset * 56;
                    auto Max = License_Infos[License_Map[i].Pos].Size - Offset;
                    if (Max > 56)
                        Max = 56;
                    for (uint8_t j = 0; j < Max; j++)
                        if (Value & ((uint64_t)1 << j))
                            License->SetSupported(License_Map[i].Pos, License_Map[i].List[Offset + j]);
                }
            }
        }
    }
    // Place holder for license key decoding - End
}

//***************************************************************************
// license_internal
//***************************************************************************

//---------------------------------------------------------------------------
license_internal::license_internal(bool IgnoreDefault)
{
    // Default license flags
    if (!IgnoreDefault)
        for (size_t i = 0; DefaultLicense_Parsers[i].Value != (uint8_t)-1; i++)
            SetSupported(DefaultLicense_Parsers[i].Value, DefaultLicense_Parsers[i].Flavor);
}

size_t license_internal::Flags_Pos_Get(uint8_t Type, uint8_t SubType)
{
    size_t Flags_Pos = 0;
    for (size_t i = 0; i < Type; i++)
        Flags_Pos += License_Infos[i].Size;
    Flags_Pos += SubType;

    if (Flags_Pos >= Flags_.size())
        Flags_.resize(Flags_Pos + 1);

    return Flags_Pos;
}

void license_internal::SetSupported(uint8_t Type, uint8_t SubType)
{
    Flags_[Flags_Pos_Get(Type, SubType)] = true;
}

bool license_internal::IsSupported(uint8_t Type, uint8_t SubType)
{
    return Flags_[Flags_Pos_Get(Type, SubType)];
}

//---------------------------------------------------------------------------
bool license_internal::FromString(const string& FromUser)
{
    // SizeOnDisk
    if (FromUser.size() > 200)
    {
        cerr << "Invalid size of license key." << endl;
        return true;
    }

    // Analyze
    uint8_t Buffer[100];
    size_t Buffer_Offset = 0;
    for (; Buffer_Offset < FromUser.size() / 2; Buffer_Offset++)
    {
        const char L1 = FromUser[Buffer_Offset * 2];
        const char L2 = FromUser[Buffer_Offset * 2 + 1];
        if (!((L1 >= '0' && L1 <= '9') || (L1 >= 'A' && L1 <= 'F'))
            && !((L2 >= '0' && L2 <= '9') || (L2 >= 'A' && L2 <= 'F')))
        {
            cerr << "Invalid characters in license key." << endl;
            return true;
        }

        Buffer[Buffer_Offset] = ((L1 - ((L1 <= '9') ? '0' : ('A' - 10))) << 4)
            | ((L2 - ((L2 <= '9') ? '0' : ('A' - 10))));
    }

    // Parse
    return FromBuffer(buffer_view(Buffer, Buffer_Offset));
}

//---------------------------------------------------------------------------
bool license_internal::FromBuffer(const buffer_view& Buffer)
{
    license_input Input;
    Input.License = this;
    if (Input.Parse(Buffer))
    {
        if (!Input.IsSupported())
        {
            cerr << "License detected but decoding not supported in this version." << endl;
            *this = license_internal();
            return false;
        }
        // TEMP IsSupported !License
        return true;
    }
    return false;
}

//---------------------------------------------------------------------------
string license_internal::ToString()
{
    // To string
    auto Buffer = ToBuffer();
    string ToReturn;
    ToReturn.reserve(Buffer.Size());
    for (size_t i = 0; i < Buffer.Size(); i++)
    {
        uint8_t Value = Buffer[i];
        Value >>= 4;
        Value += (Value < 10 ? '0' : ('A' - 10));
        ToReturn.push_back((char)Value);
        Value = Buffer[i];
        Value &= 0xF;
        Value += (Value < 10 ? '0' : ('A' - 10));
        ToReturn.push_back((char)Value);
    }
    return ToReturn;
}

//***************************************************************************
// License class
//***************************************************************************

//---------------------------------------------------------------------------
license::license()
{
    Internal = new license_internal();
}

//---------------------------------------------------------------------------
license::~license()
{
    delete (license_internal*)Internal;
}

//---------------------------------------------------------------------------
void license::Feature(feature Value) { license_internal* License = (license_internal*)Internal; License->Input_Flags[0] |= ((uint64_t)1 << (int)Value); }
void license::Muxer(muxer Value) { license_internal* License = (license_internal*)Internal; License->Input_Flags[1] |= ((uint64_t)1 << (int)Value); }
void license::Encoder(encoder Value) { license_internal* License = (license_internal*)Internal; License->Input_Flags[2] |= ((uint64_t)1 << (int)Value); }

//---------------------------------------------------------------------------
bool license::LoadLicense(string LicenseKey, bool StoreLicenseKey)
{
    // Read license from environment variable
    if (LicenseKey.empty())
        LicenseKey = getenv_string("RAWcooked_License");

    // Read license from a file
    if (LicenseKey.empty())
    {
        string Path = GetLocalConfigPath() + PathSeparator "Config.txt";
        ifstream F(Path);
        string temp;
        while (!F.fail() && !F.eof())
        {
            F >> temp;
            if (temp.rfind("License=", 8) == 0 || temp.rfind("Licence=", 8) == 0)
                LicenseKey = temp.substr(8);
        }
    }

    // Decode
    if (!LicenseKey.empty())
    {
        license_internal* License = (license_internal*)Internal;
        if (License->FromString(LicenseKey))
            return true;
        
        if (StoreLicenseKey)
        {
            string Path = GetLocalConfigPath(true);
            Path += PathSeparator "Config.txt";
            fstream F(Path, std::fstream::in | std::fstream::out);
            if (!F.is_open())
            {
                F.open(Path, ios_base::app);
                if (!F.is_open())
                {
                    cerr << "Can not open " << Path << '.' << endl;
                    return true;
                }
            }
            else
            {
                F.seekp(-1, ios_base::end);
                if (!F.fail())
                {
                    char temp;
                    F.read(&temp, 1);
                    if (F.fail())
                    {
                        cerr << "Can not read in " << Path << '.' << endl;
                        return true;
                    }
                    if (!F.fail() && !F.eof() && temp != '\n')
                    {
                        F.seekp(0, ios_base::end);
                        F.write("\n", 1); // endl;
                        if (F.fail())
                        {
                            cerr << "Can not write in " << Path << '.' << endl;
                            return true;
                        }
                    }
                    else
                        F.seekp(0, ios_base::end);
                }
            }
            F.write(string("License=" + LicenseKey + '\n').c_str(), 9 + LicenseKey.size());
            if (F.fail())
            {
                cerr << "Can not write in " << Path << '.' << endl;
                return true;
            }
        }
    }

    return false;
}

//---------------------------------------------------------------------------
bool license::ShowLicense(bool Verbose, uint64_t NewSublicenseId, uint64_t NewSubLicenseDur)
{
    license_internal* License = (license_internal*)Internal;

    // UserID
    if (License->OwnerID)
    {
        if (Verbose)
        {
            cerr << "License owner ID = " << License->OwnerID;
            if (License->SubLicenseeID)
                cerr << " (sub-licensee ID = " << License->SubLicenseeID << ')';
            cerr << '.';
            if (License->Date == (time_t)-1)
                cerr << '\n' << endl;
            else
                cerr << ' ';
        }
    }
    else if (License->Date == (time_t)-1)
        cerr << "Using default license.\nConsider to support RAWcooked: https://MediaArea.net/RAWcooked\n" << endl;

    // Date
    if (License->Date != (time_t)-1)
    {
        struct tm TimeInfo = localtime_tm(License->Date);
        cerr << "Temporary license, valid until " << TimeInfo.tm_year + 1900 << '-' << setfill('0') << setw(2) << TimeInfo.tm_mon + 1 << '-' << setfill('0') << setw(2) << TimeInfo.tm_mday << ".\n" ;

        time_t Time;
        time(&Time);
        if (Time > License->Date)
        {
            cerr << "Outdated license, using default license.\n";
            auto SubLicenseeID_Sav = License->SubLicenseeID;
            *License = license_internal();
            License->SubLicenseeID = SubLicenseeID_Sav;
        }

        cerr << endl;
    }

    if (NewSublicenseId)
    {
        if (License->SubLicenseeID)
        {
            cerr << "Error: a sublicensee can not sublicense.\n";
            return true;
        }
        License->SubLicenseeID = NewSublicenseId;
        License->SetDateFromMonths((int)NewSubLicenseDur);

        cerr << "Sub-licensee license key: ";
        cout << License->ToString();
        cerr << endl;
    }

    if (!Verbose)
        return false;

    // Info
    for (int8_t i = 0; i < License_Infos_Size; i++)
    {
        cerr << "Licensed " << License_Infos[i].Name << ':' << endl;
        for (int8_t j = 0; j < License_Infos[i].Size; j++)
        {
            const char* Supported = License->IsSupported(i, j) ? "Yes" : "No ";
            cerr << Supported << ' ' << License_Infos[i].Flavor_String(j) << endl;
        }
        cerr << endl;
    }

    return false;
}

//---------------------------------------------------------------------------
bool license::IsSupported()
{
    license_internal* License = (license_internal*)Internal;

    // Info
    for (int8_t i = 0; i < License_Parser_Offset; i++)
        for (int8_t j = 0; j < License_Infos[i].Size; j++)
            if (License->Input_Flags[i] & ((uint64_t)1 << j) && !License->IsSupported(i, j))
                return false;
    
    return true;
}

//---------------------------------------------------------------------------
bool license::IsSupported(feature Feature)
{
    license_internal* License = (license_internal*)Internal;

    return License->IsSupported(0, (uint8_t)Feature);
}

//---------------------------------------------------------------------------
bool license::IsSupported(parser Parser, uint8_t Flavor)
{
    license_internal* License = (license_internal*)Internal;

    return License->IsSupported(Parser, Flavor);
}

//---------------------------------------------------------------------------
bool license::IsSupported_License()
{
    license_input Input;
    buffer Buffer;
    return !Input.Parse(Buffer) && Input.IsSupported();
}

//---------------------------------------------------------------------------
void license_internal::SetDateFromMonths(int Months)
{
    time_t Time;
    time(&Time);
    struct tm TimeInfo = localtime_tm(Time);
    TimeInfo.tm_mon += (decltype(TimeInfo.tm_mon))Months;
    Time = mktime(&TimeInfo);
    if (Date < 0 || Date > Time)
        Date = Time;
}
